import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;

import javafx.scene.image.Image;
import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;

/**
 * マテリアルリーダー
 * <pre>
 * </pre>
 */
public class MaterialReader {

	/**
	 * デバッグレベル
	 * <pre>
	 * 通常は0にしてください。
	 * １で概要レベル、２で詳細レベルまで出力します。
	 * </pre>
	 */
	private static int			lvlDebug			= 0;

	/**
	 * マテリアル取得
	 * @param strFilePath ファイルパス
	 * @return HashMap<String, PhongMaterial> マテリアル名毎のマテリアルを返却します。
	 */
	public static HashMap<String, PhongMaterial> getMaterials(String strFilePath) throws Exception {

		// デフォルトパス
		String				strDefaultPath	= null;

		// InputStream
		InputStream			is				= null;

		// InputStreamReader
		InputStreamReader	isr				= null;

		// LineNumberReade
		LineNumberReader	lnr				= null;

		// 行（エラー表示用）
		int row = 0;

		try{

			// デバッグ情報出力
			outputDebugInfo(1, "START MaterialReader. strFilePath:" + strFilePath);

			// ファイルパスのデフォルトパスを保持
			int indexDefaultPath = strFilePath.lastIndexOf("/");
			if(indexDefaultPath == -1){
				strDefaultPath = "";
			}else{
				strDefaultPath = strFilePath.substring(0, indexDefaultPath + 1);
			}

			// InputStreamを生成
			is = ObjReader.application.getClass().getResourceAsStream(strFilePath);

			// ファイルが見つからない場合、
			if(is == null){
				// エラーをスロー
				throw new IllegalStateException("File Not Found. filePath = " + strFilePath);
			}

			// InputStreamReaderを生成
			isr = new InputStreamReader(is);

			// LineNumberReaderを生成
			lnr = new LineNumberReader(isr);

			// マテリアルマップ生成
			HashMap<String, PhongMaterial> mapMaterial = new HashMap<String, PhongMaterial>();

			// マテリアル名
			String strMaterialName = null;

			// マテリアル
			PhongMaterial material = null;

			// 行数分、ループ
			while(true){

				// 行＋１
				row++;

				// １行取得
				String strLine = lnr.readLine();

				// 行情報が取得できない場合、
				if(strLine == null){
					// ループを抜ける
					break;
				}

				// トークンを生成
				StringTokenizer stLine = new StringTokenizer(strLine, " ");

				// 文字列一覧生成
				ArrayList<String> strLst = new ArrayList<String>(5);

				// トークン分、ループ
				while(true){

					// トークンがない場合、
					if(!stLine.hasMoreTokens()){
						// ループを抜ける
						break;
					}

					// トークンを文字列一覧に格納
					strLst.add(stLine.nextToken());

				} // end while トークン分、ループ

				// 文字列が取得できなかった場合、
				if(strLst.size() == 0){
					// コンティニュー
					continue;
				}

				//
				// 各情報を取得
				//

				// マテリアルの場合、
				if(strLst.get(0).equals("newmtl")){

					// マテリアル名が既に存在する場合、
					if(strMaterialName != null){

						// マテリアル設定を保存する
						mapMaterial.put(strMaterialName, material);

					}

					// マテリアル名を保存
					strMaterialName = strLst.get(1);

					// マテリアルを生成
					material = new PhongMaterial();

				// スペキュラ指数の場合、
				}else if(strLst.get(0).equals("Ns")){

					// スペキュラ指数を設定
					material.setSpecularPower(Double.parseDouble(strLst.get(1)));

					// デバッグ情報出力
					outputDebugInfo(1, "Ns=" + strLst.get(1));

				// スペキュラ色の場合、
				// 鏡面反射
				}else if(strLst.get(0).equals("Ks")){

					// カラーを取得
					Color color = MaterialReader.getColor(strLst);

					// スペキュラ色（鏡面反射）を設定
					material.setSpecularColor(color);

					// デバッグ情報出力
					outputDebugInfo(1, "Ks=" + color);

				// アンビエント色の場合、
				// 環境光
				}else if(strLst.get(0).equals("Ka")){

				// ディフューズ色の場合、
				// 拡散反射
				}else if(strLst.get(0).equals("Kd")){

					// カラーを取得
					Color color = MaterialReader.getColor(strLst);

					// ディフューズ色（拡散反射）を設定
					material.setDiffuseColor(color);

					// デバッグ情報出力
					outputDebugInfo(1, "Kd=" + color);

				// バンプマップの場合、
				}else if(strLst.get(0).equals("map_Bump")){

					// イメージを取得
					Image image = MaterialReader.getImage(strDefaultPath, strLst);

					// バンプマップを設定
					material.setBumpMap(image);

				// ディフューズテクスチャマップの場合、
				// 拡散マップ
				}else if(strLst.get(0).equals("map_Kd")){

					// イメージを取得
					Image image = MaterialReader.getImage(strDefaultPath, strLst);

					// ディフューズテクスチャマップ（拡散マップ）を設定
					material.setDiffuseMap(image);

				// 自己照明マップの場合、
				}else if(strLst.get(0).equals("???")){

					// イメージを取得
					Image image = MaterialReader.getImage(strDefaultPath, strLst);

					// 自己照明マップを設定
					material.setSelfIlluminationMap(image);

				// スペキュラカラーテクスチャマップ
				// 鏡面マップの場合、
				}else if(strLst.get(0).equals("map_Ks")){

					// イメージを取得
					Image image = MaterialReader.getImage(strDefaultPath, strLst);

					// スペキュラカラーテクスチャマップ（鏡面マップ）を設定
					material.setSpecularMap(image);

				} // end if

			} // end while 行数分、ループ

			// マテリアル名が既に存在する場合、
			if(strMaterialName != null){

				// マテリアル設定を保存する
				mapMaterial.put(strMaterialName, material);

			}

			// ひとつ以上、マテリアル設定がある場合、
			if(mapMaterial.size() >= 1){

				// リターン
				return mapMaterial;

			}

			// デバッグ情報出力
			outputDebugInfo(1, "END MaterialReader. strFilePath:" + strFilePath);

			// リターン
			return null;

		}catch(Exception ex){

			// 行が途中の場合、
			if(row > 0){
				// エラーのあった行を出力
				System.out.println("Error! row = : " + row);
			}

			// そのままエラーをスロー
			throw ex;

		}finally{

			try{
				// クローズ
				if(isr != null){
					isr.close();
				}
			}catch(Exception ex){
				ex.printStackTrace();
			}

		}

	} // end getMaterials

	/**
	 * カラー取得
	 * <pre>
	 * テキストデータからカラーオブジェクトを取得します。
	 * </pre>
	 * @param strLst 文字列一覧
	 * @return Color カラーオブジェクト
	 */
	private static Color getColor(ArrayList<String> strLst){

		Color color = Color.color(Double.parseDouble(strLst.get(1)),
									Double.parseDouble(strLst.get(2)),
									Double.parseDouble(strLst.get(3))
		);

		return color;

	} // end getColor

	/**
	 * イメージ取得
	 * <pre>
	 * ファイルパスからイメージを取得します。
	 * BMP/GIF/JPEG/PNGを対象とします。
	 * </pre>
	 * @param strDefaultPath デフォルトパス
	 * @param strLst 文字列一覧
	 * @return Image イメージオブジェクト
	 */
	private static Image getImage(String strDefaultPath, ArrayList<String> strLst){

		// デバッグ情報出力
		outputDebugInfo(1, "START getImage.");

		// 文字列一覧の件数を取得
		int countStrLst = strLst.size();

		// 文字列一覧の件数分、ループ
		for(int i = 1; i < countStrLst; i++){

			// ファイル名（候補）を取得
			String strPath = strLst.get(i);

			// 拡張子の位置を取得
			int index = strPath.lastIndexOf(".");

			// 拡張子の位置が取得できた場合、
			if(index != -1){

				// 拡張子を取得し、大文字に変換
				String sttExt = strPath.substring(index + 1).toUpperCase();;

				// 拡張子が対象の拡張子の場合、
				if(sttExt.equals("BMP") || sttExt.equals("GIF") || sttExt.equals("PNG") ||
					sttExt.equals("JPG") || sttExt.equals("JPEG") || sttExt.equals("JPE") || sttExt.equals("JFIF")){

					// URIパスを取得
					String strURIPath = new File(strDefaultPath + strPath).toURI().toString();

					// イメージを取得
					Image image = new Image(strURIPath);

					// デバッグ情報出力
					outputDebugInfo(2, "strURIPath:" + strURIPath);
					outputDebugInfo(2, "image:" + image);
					outputDebugInfo(2, "isError:" + image.isError());

					return image;

				} // end if 拡張子が対象の拡張子の場合

			} // end if 拡張子の位置が取得できた場合

		} // end for 文字列一覧の件数分、ループ

		// デバッグ情報出力
		outputDebugInfo(1, "END getImage.");

		// nullでリターン
		return null;

	} // end getImage

	/**
	 * デバッグ情報出力
	 * <pre>
	 * デバッグレベルをチェックし、コンソールに情報を出力します。
	 * </pre>
	 * @param strDebugInfo デバッグ情報
	 */
	private static void outputDebugInfo(int lvlDebug, String strDebugInfo){

		if(lvlDebug <= MaterialReader.lvlDebug){

			System.out.println(strDebugInfo);

		}

	} // end outputDebugInfo

}
