import java.util.ArrayList;
import java.util.HashMap;

import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.application.Application;
import javafx.application.Platform;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.ObservableList;
import javafx.event.Event;
import javafx.scene.AmbientLight;
import javafx.scene.Group;
import javafx.scene.LightBase;
import javafx.scene.Node;
import javafx.scene.PerspectiveCamera;
import javafx.scene.PointLight;
import javafx.scene.Scene;
import javafx.scene.SubScene;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.ButtonType;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.layout.StackPane;
import javafx.scene.paint.Color;
import javafx.scene.paint.CycleMethod;
import javafx.scene.paint.LinearGradient;
import javafx.scene.paint.Stop;
import javafx.scene.text.Font;
import javafx.scene.text.FontWeight;
import javafx.scene.text.Text;
import javafx.scene.transform.Translate;
import javafx.stage.Stage;
import javafx.util.Duration;

/**
 * アプリケーションクラス
 */
public class PRApplication extends Application implements PRConstant {

	/** フェーズ */
	private		int						phase					= PHASE_INIT;

	/** 時間 */
	private		int						timeCtrl				= 0;
	private		int						time					= 0;

	/** プレイヤー数 */
	private		int						countPrusher			= 0;

	/** ステージ */
	private		int						stage					= 0;

	/** スコア */
	private		int						score					= 0;

	/** ハイスコア */
	private		int						scoreHigh				= 30000;

	/** モード */
	private		int						mode					= 0;

	/** ペイン */
	private		StackPane				pane					= null;

	/** シーン */
	private		Scene					scene					= null;

	/** 3Dシーン */
	private		SubScene				scene3D					= null;

	/** 2Dシーン */
	private		SubScene				scene2D					= null;

	/** カメラ */
	private		PerspectiveCamera		camera					= null;

	/** 光源 */
	private		LightBase				light					= null;

	/** パネル3D */
	private		PRPanel3D				panel3D					= null;

	/** パネル2D */
	private		Group					panel2D					= null;

	/** テキスト */
	private		Text					textKey					= null;
	private		Text					textControlStick		= null;
	private		Text					textReverse				= null;
	private		Text					textStage				= null;
	private		Text					textTime				= null;
	private		Text					textScore				= null;
	private		Text					textScoreHigh			= null;
	private		Text					textPrusher				= null;
	private		Text					textBy					= null;
	private		Text					textStage2				= null;
	private		Text					textGameOver			= null;

	/** キーを押している状態一覧 */
	private HashMap<KeyCode, String>	mapKeyPress				= null;

	// ゲーム中フラグ
	private		boolean					flgPlaying				= true;

	// タイトル
	private		PRSpriteTitlePrusher	spriteTitlePrusher		= null;

	// タイトルアステロイダー
	private		PRSpriteTitleShip		spriteTitleShip			= null;

	/** モードセレクト */
	private		PRSpriteMode			spriteMode				= null;

	/** プレイヤー */
	private		PRSpritePrusher			spritePrusher			= null;

	/** ロック */
	private		PRLock					lock					= null;

	/** 地面 */
	private		PRSpriteGround			spriteGround			= null;

	/** 空 */
	private		PRSpriteSky				spriteSky				= null;

	/** タイル */
	private		PRSpriteTile			spriteTile				= null;

	/** スプライトグループ雲 */
	private		PRSpriteGroupCloud		spriteGroupCloud		= null;

	/** 効果音管理 */
	private		int						indexShot				= 0;
	private		int						indexBomb				= 0;
	private		int						indexHew				= 0;
	private		int						indexRapel				= 0;
	private		int						indexTamaTeki			= 0;
	private		int						indexLaser1Teki			= 0;
	private		int						indexLockon				= 0;

	/** BGM／効果音 */
	private		PRSoundMP3				mp3Title				= null;
	private		PRSoundMP3				mp3BGM					= null;
	private		PRSoundMP3				mp3BGMTengu				= null;
	private		PRSoundMP3				mp3BGMRedEye			= null;
	private		PRSoundMP3				mp3Talk1				= null;
	private		PRSoundMP3				mp3Talk2				= null;
	private		PRSoundMP3				mp3Ending				= null;
	private		PRSoundEffect[]			waveShots				= null;
	private		PRSoundEffect[]			waveBombs				= null;
	private		PRSoundEffect[]			waveHews				= null;
	private		PRSoundEffect[]			waveRapels				= null;
	private		PRSoundEffect[]			waveTamaTekis			= null;
	private		PRSoundEffect[]			waveLaser1Tekis			= null;
	private		PRSoundEffect[]			waveLockons				= null;
	private		PRSoundMP3				mp3Bump					= null;
	private		PRSoundMP3				mp3BGMFanfare			= null;
	private		PRSoundMP3				mp3GameOver				= null;

	/** フェードアウトフラグ */
	private		boolean					flgFadeout				= false;
	private		boolean					flgFadeoutFanfare		= false;

	/** フェードアウトボリューム */
	private		double					volumeFadeout			= 0.02;

	/** ボリューム（BGM） */
	private		double					volume					= 1.0;
	private		double					volumeFanfare			= 1.0;

	/** ストーリー */
	private		PRStoryPrusher			storyPrusher			= null;

	/** スプライトグループ一覧 */
	private		ArrayList<PRSpriteGroup>	spriteGroupLst			= null;

	/**
	 * 実行
	 * <pre>
	 * 当メソッドから、はじまります。
	 * </pre>
	 * @param args 引数
	 */
	public static void main(String[] args){

		// アプリケーションを起動します
		launch(args);

	} // end main

	/**
	 * スタート
	 * <pre>
	 * 初期化を行います。
	 * </pre>
	 */
	@Override
	public void start(Stage primaryStage){

		try{

			// PRImageにアプリケーションを設定
			PRImage.setApplication(this);

			// ObjReaderにアプリケーションを設定
			ObjReader.setApplication(this);

			// スプライト初期化
			PRSpriteMode.init();
			PRSpriteSky.init();
			PRSpriteTitlePrusher.init();
			PRSpriteTitleShip.init();
			PRSpritePrusher.init();
			PRSpriteTama.init();
			PRSpriteRock.init();
			PRSpriteTengu.init();
			PRSpriteTree.init();
			PRSpritePole.init();
			PRSpriteBuilding.init();
			PRSpriteStone.init();
			PRSpriteShrub.init();
			PRSpriteCloud.init();
			PRSpriteShip.init();
			PRSpriteRedEye.init();
			PRSpriteTamaTeki.init();
			PRSpriteEnding.init();

			// ペインを生成
			pane = new StackPane();

			// シーンを生成
			// Zバッファを有効にする（最後の引数、boolean）
			scene = new Scene(pane, PRConstant.WIDTH, PRConstant.HEIGHT, true);

			// 背景を黒にする
			scene.setFill(Color.BLACK);

			// シーンを設定
			primaryStage.setScene(scene);

			// タイトル設定
			primaryStage.setTitle("Prusher3D");

			// ウィンドウを閉じる際に安全に終了する
			primaryStage.setOnCloseRequest(req -> {
				// JavaFXアプリケーションを終了する
				Platform.exit();
				// 他に終了していないスレッドを終了する
				System.exit(0);
			});

			// BGM・効果音を生成
			mp3Title = new PRSoundMP3(this, "sound/PRMusic_Opening.mp3", true);
			mp3BGM = new PRSoundMP3(this, "sound/PRMusic_MainBGM.mp3", true);
			mp3BGMTengu = new PRSoundMP3(this, "sound/PRMusic_TENGU.mp3", true);
			mp3BGMRedEye = new PRSoundMP3(this, "sound/PRMusic_RedEyes.mp3", true);
			mp3Talk1 = new PRSoundMP3(this, "sound/PRVo_GetReady.mp3", false);
			mp3Talk2 = new PRSoundMP3(this, "sound/PRVo_YourDoingGreat.wav", false);
			mp3Ending = new PRSoundMP3(this, "sound/PRMusic_Ending.mp3", false);
			waveShots = new PRSoundEffect[10];
			for(int i = 0; i < 10; i++){
				waveShots[i] = new PRSoundEffect(this, "sound/PRSE_Shot.wav", false);
			}
			waveBombs = new PRSoundEffect[10];
			for(int i = 0; i < 10; i++){
				waveBombs[i] = new PRSoundEffect(this, "sound/PRSE_Bom1.wav", false);
			}
			waveHews = new PRSoundEffect[10];
			for(int i = 0; i < 10; i++){
				waveHews[i] = new PRSoundEffect(this, "sound/PRSE_Bom2.wav", false);
			}
			waveRapels = new PRSoundEffect[10];
			for(int i = 0; i < 10; i++){
				waveRapels[i] = new PRSoundEffect(this, "sound/PRSE_Reflection.wav", false);
			}
			waveTamaTekis = new PRSoundEffect[10];
			for(int i = 0; i < 10; i++){
				waveTamaTekis[i] = new PRSoundEffect(this, "sound/PRSE_Lase1.wav", false);
			}
			waveLaser1Tekis = new PRSoundEffect[10];
			for(int i = 0; i < 10; i++){
				waveLaser1Tekis[i] = new PRSoundEffect(this, "sound/PRSE_Lase2.wav", false);
			}
			waveLockons = new PRSoundEffect[10];
			for(int i = 0; i < 10; i++){
				waveLockons[i] = new PRSoundEffect(this, "sound/PRSE_LockOn.wav", false);
			}
			mp3Bump = new PRSoundMP3(this, "sound/PRVo_Ah.mp3", false);
			mp3BGMFanfare = new PRSoundMP3(this, "sound/PRMusic_Fanfare.mp3", false);
			mp3GameOver = new PRSoundMP3(this, "sound/PRMusic_GameOver.mp3", false);

			// ストーリークラスを生成
			storyPrusher = new PRStoryPrusher(this);

			// スプライトグループ一覧生成
			spriteGroupLst = new ArrayList<PRSpriteGroup>(100);

			// グループを生成
			panel3D = new PRPanel3D(this);

			// スプライト一覧を生成
			scene3D = new SubScene(panel3D, PRConstant.WIDTH, PRConstant.HEIGHT, true, null);

			// スプライト層一覧に追加
			pane.getChildren().add(scene3D);

			// カメラ設定
			initCamera();

			// 光源設定
			initLight();

			// アンビエント光
			AmbientLight ambientLight = new AmbientLight(Color.rgb(60, 60, 60));
			panel3D.getChildren().add(ambientLight);

			// 地面を作成
			spriteGround = new PRSpriteGround(this);

			// 地面を追加
			panel3D.getChildren().add(spriteGround);

			// 空を作成
			spriteSky = new PRSpriteSky(this);

			// 空を追加
			panel3D.getChildren().add(spriteSky);

			// タイルを作成
			spriteTile = new PRSpriteTile(this);

			// タイルを追加
			panel3D.getChildren().add(spriteTile);

			// グループを生成
			panel2D = new Group();

			// スプライト一覧を生成
			scene2D = new SubScene(panel2D, PRConstant.WIDTH, PRConstant.HEIGHT);

			// スプライト層一覧に追加
			pane.getChildren().add(scene2D);

			// テキストを生成
			textKey = new Text(272, 500, "PUSH SPACE KEY");
			textControlStick = new Text(320, 550, "CONTROL STICK MODE");
			textReverse = new Text(320, 580, "REVERSE MODE");
			textStage = new Text(640, 20, "STAGE:" + stage);
			textTime = new Text(640, 792, "TIME:" + storyPrusher.getStoryTime());
			textScore = new Text(64, 20, "SCORE:" + score);
			textScoreHigh = new Text(288, 20, "HIGH-SCORE:" + scoreHigh);
			textPrusher = new Text(32, 792, "PRUSHER: ■■■■■");
			textBy = new Text(204, 720, "PROGRAMED  BY  KIN  /  MUSIC  BY  Aki2");
			textStage2 = new Text(304, 384, "STAGE");
			textGameOver = new Text(304, 400, "GAME OVER");

			// フォント設定
			textKey.setFont(Font.font("Meiryo", FontWeight.BOLD, 32));
			textControlStick.setFont(Font.font("Meiryo", FontWeight.BOLD, 20));
			textReverse.setFont(Font.font("Meiryo", FontWeight.BOLD, 20));
			textStage.setFont(Font.font("Meiryo", FontWeight.BOLD, 20));
			textTime.setFont(Font.font("Meiryo", FontWeight.BOLD, 20));
			textScore.setFont(Font.font("Meiryo", FontWeight.BOLD, 20));
			textScoreHigh.setFont(Font.font("Meiryo", FontWeight.BOLD, 20));
			textPrusher.setFont(Font.font("Meiryo", FontWeight.BOLD, 20));
			textBy.setFont(Font.font("Meiryo", FontWeight.BOLD, 20));
			textStage2.setFont(Font.font("Meiryo", FontWeight.BOLD, 48));
			textGameOver.setFont(Font.font("Meiryo", FontWeight.BOLD, 32));

			Stop[] stops = new Stop[] { new Stop(0, Color.BLUE), new Stop(0.5, Color.WHITE), new Stop(1, Color.BLUE)};
			LinearGradient lgStage2 = new LinearGradient(1, 0, 0, 1, true, CycleMethod.NO_CYCLE, stops);

			// 色の設定
			textKey.setFill(Color.WHITE);
			textControlStick.setFill(Color.YELLOW);
			textReverse.setFill(Color.LIGHTGRAY);
			textStage.setFill(Color.LIGHTGRAY);
			textTime.setFill(Color.GRAY);
			textScore.setFill(Color.WHITE);
			textScoreHigh.setFill(Color.BLUE);
			textPrusher.setFill(Color.YELLOW);
			textBy.setFill(Color.WHITE);
			textStage2.setFill(lgStage2);
			textGameOver.setFill(Color.WHITE);

			textKey.setStroke(Color.BLACK);
			textControlStick.setStroke(Color.BLACK);
			textReverse.setStroke(Color.BLACK);
			textStage.setStroke(Color.BLACK);
			textTime.setStroke(Color.BLACK);
			textScore.setStroke(Color.BLACK);
			textScoreHigh.setStroke(Color.WHITE);
			textPrusher.setStroke(Color.RED);
			textBy.setStroke(Color.BLUE);
			textStage2.setStroke(Color.BLACK);
			textGameOver.setStroke(Color.BLACK);

			// テキストを追加
			panel2D.getChildren().add(textStage);
			panel2D.getChildren().add(textTime);
			panel2D.getChildren().add(textScore);
			panel2D.getChildren().add(textScoreHigh);
			panel2D.getChildren().add(textPrusher);

			// タイトルを生成
			spriteTitlePrusher = new PRSpriteTitlePrusher(this);

			// タイトルアステロイダーを生成
			spriteTitleShip = new PRSpriteTitleShip(this);

			// プレイヤーを生成
			spritePrusher = new PRSpritePrusher(this);

			// ロックを生成
			lock = new PRLock(this);

			// キー押下状況一覧を生成
			mapKeyPress = new HashMap<KeyCode, String>();

			// キーを押したイベントを登録
			scene.addEventHandler(KeyEvent.KEY_PRESSED, event -> sceneKeyPressed(event));

			// キーを離したイベントを登録
			scene.addEventHandler(KeyEvent.KEY_RELEASED, event -> sceneKeyReleased(event));

			// ウインドウサイズ変更リスナー追加
			pane.widthProperty().addListener((observable, oldValue, newValue) -> windowResized());
			pane.heightProperty().addListener((observable, oldValue, newValue) -> windowResized());

			// ウィンドウ表示
			primaryStage.show();

			// 時間初期化
			timeCtrl = 0;
			time = 0;

			// タイマーをスタート
			Timeline timer = new Timeline(new KeyFrame(Duration.millis(4), event -> runCtrl()));
			timer.setCycleCount(Timeline.INDEFINITE);
			timer.play();

		}catch(Exception ex){
			// スタックトレース出力
			ex.printStackTrace();
			// エラーを表示
			Alert alert = new Alert(AlertType.ERROR , "" , ButtonType.OK);
			alert.setTitle("エラー");
			alert.getDialogPane().setContentText(ex.toString());
			alert.showAndWait();
		}

	} // end start

	/**
	 * タイマー実行
	 * <pre>
	 * タイマーから定期的に呼ばれます。
	 * </pre>
	 */
	synchronized public void runCtrl() {

		// 時間＋１
		timeCtrl++;

		// 制御の時間
		if(timeCtrl % 4 == 0){

			// 実行メソッド呼び出し
			this.run();

		// 各ノードの制御の時間
		}else if(timeCtrl % 4 == 1){

			// 3Dパネルを実行
			panel3D.run();

		// グループの制御／お掃除の時間
		}else if(timeCtrl % 4 == 3){

			// グループ実行メソッド呼び出し
			this.runGroup();

			// お掃除の時間：スプライトグループ
			if(timeCtrl % 40 == 3){

				// グループクリーンメソッド呼び出し
				this.cleanGroup();

			// クリーン
			}else if(timeCtrl % 40 == 11){

				// 3Dパネルのクリーンメソッドを実行
				panel3D.clean();

			} // end if

			// フェーズ−ゲーム中の場合、
			if(phase == PHASE_GAME){

				// ロックオン
				if(timeCtrl % 40 == 19){

					// ロックオン
					Platform.runLater( () -> lock.lockon());

				// ロックオン外す
				}else if(timeCtrl % 40 == 27){

					// ロックオン外す
					Platform.runLater( () -> lock.unlock());

				} // end if 時間

			} // end if フェーズ−ゲーム中の場合

		} // end if 時間

	} // end runCtrl

	/**
	 * 実行
	 * <pre>
	 * タイマーから定期的に呼ばれます。
	 * </pre>
	 */
	synchronized public void run() {

		// 時間＋１
		time++;

		// フェーズ−初期化の場合、
		if(phase == PHASE_INIT){

			// ステージ初期化
			stage = 0;

			// スコア初期化
			score = 0;

			// 時間初期化
			time = 0;

			// プレイヤー数
			countPrusher = 5;

			// スプライトグループの要素をクリア
			spriteGroupLst.clear();

			// ストーリーをリセット
			storyPrusher.reset();

			// カメラ位置を戻す
			camera.setTranslateX(0.0);
			camera.setTranslateY(0.0);

			// タイトルを初期化
			spriteTitlePrusher.initPosition();

			// タイトルを追加
			this.addNode(spriteTitleShip);

			// フェーズ−タイトル１へ
			phase = PHASE_TITLE_1;

		// フェーズ−タイトル１の場合、
		}else if(phase == PHASE_TITLE_1){

			// ３秒
			if(time == 180){

				// タイトルを追加
				this.addNode(spriteTitlePrusher);

			}

			// スペースが押された場合、
			if(this.isKeyPressed(KeyCode.SPACE)){

				// ３秒未満の場合、
				if(time < 180){

					// タイトルを追加
					this.addNode(spriteTitlePrusher);

				}

				// タイトルをスキップ
				spriteTitlePrusher.skipTitle();

				// フェーズ−タイトル２へ
				phase = PHASE_TITLE_2;

			}

		// フェーズ−タイトル２の場合、
		}else if(phase == PHASE_TITLE_2){

			// タイトルから待ち

		// フェーズ−タイトル３の場合、
		}else if(phase == PHASE_TITLE_3){

			// 時間が来たら、
			if(time == 1){

				// テキストを追加
				panel2D.getChildren().add(textKey);
				panel2D.getChildren().add(textControlStick);
				panel2D.getChildren().add(textReverse);
				panel2D.getChildren().add(textBy);

				// モードセレクトを生成
				spriteMode = new PRSpriteMode(this);

				// モードセレクトを追加
				this.addNode(spriteMode);

				// BGMをスタート
				mp3Title.start();

			} // end if 時間が来たら

			// 上が押された場合、
			if(this.isKeyPressed(KeyCode.UP)){

				// モード変更
				mode = 0;

				// モード変更
				spriteMode.setMode(mode);

			// 下が押された場合、
			}else if(this.isKeyPressed(KeyCode.DOWN)){

				// モード変更
				mode = 1;

				// モード変更
				spriteMode.setMode(mode);

			// スペースが押された場合、
			}else if(this.isKeyPressed(KeyCode.SPACE)){

				// タイトルを削除
				this.removeNode(spriteTitlePrusher);

				// タイトルアステロイダーを削除
				this.removeNode(spriteTitleShip);

				// テキストを削除
				panel2D.getChildren().remove(textKey);
				panel2D.getChildren().remove(textControlStick);
				panel2D.getChildren().remove(textReverse);
				panel2D.getChildren().remove(textBy);

				// モードセレクトを削除
				this.removeNode(spriteMode);

				// BGMをストップ
				mp3Title.stop();

				// モード設定
				spritePrusher.setMode(mode);

				// プレイヤーの位置を戻す
				spritePrusher.initPos();

				// プレイヤーをスタート
				spritePrusher.start();

				// プレイヤーを追加
				this.addNode(spritePrusher);

				// 効果音をスタート
				mp3Talk1.start();

				// BGMをスタート
				mp3BGM.start();

				// 時間初期化
				time = 0;

				// フェーズ−ステージスタートへ
				phase = PHASE_STAGE_START;

			}

		// フェーズ−ステージスタートの場合、
		}else if(phase == PHASE_STAGE_START){

			// 時間が来たら、
			if(time == 1){

				// ステージ＋１
				stage++;

				// ステージ
				textStage.setText("STAGE:" + stage);

				// タイム
				textTime.setText("TIME:" + storyPrusher.getStoryTime());

				// テキスト−ステージを編集
				textStage2.setText("STAGE " + stage);

				// スコア
				textScore.setText("SCORE:" + score);

				// テキストを追加
				panel2D.getChildren().add(textStage2);

				// ステージ設定
				spriteSky.setStage(stage);
				spriteGround.setStage(stage);
				spriteTile.setStage(stage);

				// ステージ１の場合、
				if(stage == 1){

					// プレイヤーの影を追加
					this.addShadow(spritePrusher, 0.1);

				// ステージ３の場合、
				}else if(stage == 3){

					// プレイヤーの影を削除
					this.removeShadow(spritePrusher);

				}

			// ４秒、
			}else if(time == 240){

				// テキストを削除
				panel2D.getChildren().remove(textStage2);

				// 時間初期化
				time = 0;

				// フェーズ−ゲーム中へ
				phase = PHASE_GAME;

			} // end if 時間が来たら

		// フェーズ−ゲーム中の場合、
		}else if(phase == PHASE_GAME){

			// スコア＋１
			addScore(1);

			// スコア
			textScore.setText("SCORE:" + score);

			// ハイスコアを上回ったら、
			if(score > scoreHigh){

				// ハイスコア変更
				scoreHigh = score;

				// ハイスコア
				textScoreHigh.setText("HIGH-SCORE:" + scoreHigh);

			}

			// タイム
			textTime.setText("TIME:" + storyPrusher.getStoryTime());

			// ストーリーを進める
			storyPrusher.next();

		// フェーズ−ステージクリアの場合、
		}else if(phase == PHASE_STAGE_CLEAR){

			// 効果音をスタート
			mp3Talk2.start();

			// 時間初期化
			time = 0;

			// フェーズ−ステージスタートへ
			phase = PHASE_STAGE_START;

		// フェーズ−衝突の場合、
		}else if(phase == PHASE_BUMP){

			// 時間が来たら、
			if(time > 210){

				// プレイヤー数−１
				countPrusher--;

				// プレイヤー数変更
				StringBuilder sb = new StringBuilder();

				// プレイヤー
				for(int i = 0; i < countPrusher; i++){
					sb.append("■");
				}

				// プレイヤー数変更
				textPrusher.setText("PRUSHER: " + sb.toString());

				// プレイヤー数が０になった場合、
				if(countPrusher == 0){

					// BGMをストップ
					mp3BGM.stop();
					mp3BGMTengu.stop();
					mp3BGMRedEye.stop();

					// 時間初期化
					time = 0;

					// フェーズ−ゲームオーパーへ
					phase = PHASE_GAMEOVER;

				// プレイヤー数が残っている場合、
				}else{

					// プレイヤーのフェーズをスタートへ
					spritePrusher.start();

					// 効果音をスタート
					mp3Talk1.start();

					// 時間初期化
					time = 0;

					// フェーズ−ゲーム中へ
					phase = PHASE_GAME;

				} // end if

			} // end if // 時間が来たら

		// フェーズ−ゲームオーバーの場合、
		}else if(phase == PHASE_GAMEOVER){

			// 時間が来たら、
			if(time == 120){

				// プレイヤーを削除
				removeNode(spritePrusher);

				// テキストを追加
				panel2D.getChildren().add(textGameOver);

				// 効果音をスタート
				mp3GameOver.start();

			// 時間が来たら、
			}else if(time > 1080){

				// 雲を終了させる（存在有無は処理の中で判断するので、無条件で呼び出して問題ない）
				this.finishCloud();

				// テキストを削除
				panel2D.getChildren().remove(textGameOver);

				// 敵スプライトをクリア
				this.removeSpriteTs();

				// スプライトグループの要素をクリア
				spriteGroupLst.clear();

				// 時間初期化
				time = 0;

				// フェーズ−初期化へ
				phase = PHASE_INIT;

			}

		// フェーズ−ゲームクリアの場合、
		}else if(phase == PHASE_GAME_CLEAR){

			// 時間が来たら、
			if(time == 1){

				// BGMをストップ
				mp3BGM.stop();
				mp3BGMRedEye.stop();

				// プレイヤーにゲームクリアを伝える
				spritePrusher.clearGame();

			// 5秒
			}else if(time < 300){

				// 何もしない

			// 5秒
			}else if(time == 300){

				// プレイヤーを削除
				removeNode(spritePrusher);

				// カメラ位置を戻す
				camera.setTranslateX(0.0);
				camera.setTranslateY(0.0);

				// エンディングを生成
				PRSpriteEnding spriteEnding = new PRSpriteEnding(this);

				// ノード追加
				this.addNode(spriteEnding);

			// 上記以外
			}else{

				// 何もしない
				// 終了はエンディングクラス側から行う

			} // end if 時間

		} // end if フェーズ

		// フェーズ− ステージスタート または ゲーム中 または ゲームオーバー の場合、
		if(phase == PHASE_STAGE_START || phase == PHASE_GAME || phase == PHASE_GAMEOVER){

			// 進行
			progress();

		}

		// フェードアウト中の場合、
		if(flgFadeout){

			// フェードアウト
			volume = volume - volumeFadeout;
			mp3BGM.setVolume(volume);

			// ボリュームが０になった場合、
			if(volume <= 0.0){
				// フェードアウト完了
				flgFadeout = false;
				// ストップ
				mp3BGM.stop();
				// ボリュームを戻す
				volume = 1.0;
				Platform.runLater( () -> mp3BGM.setVolume(1.0));
			}

		} // end if フェードアウト中の場合

		// ファンファーレ フェードアウト中の場合、
		if(flgFadeoutFanfare){

			// フェードアウト
			volumeFanfare = volumeFanfare - 1.0 / 360;
			mp3BGMFanfare.setVolume(volumeFanfare);

			// ボリュームが０になった場合、
			if(volumeFanfare <= 0.0){
				// フェードアウト完了
				flgFadeoutFanfare = false;
				// ストップ
				mp3BGMFanfare.stop();
				// ボリュームを戻す
				volumeFanfare = 1.0;
				Platform.runLater( () -> mp3BGMFanfare.setVolume(1.0));
			}

		} // end if フェードアウト中の場合

	} // end run

	/**
	 * グループ実行
	 * <pre>
	 * タイマーから定期的に呼ばれます。
	 * </pre>
	 */
	synchronized public void runGroup() {

		// フェーズ−ゲーム中の場合、
		if(phase == PHASE_GAME){

			// スプライトグループ一覧の件数を取得
			int countSpriteGroupLst = spriteGroupLst.size();

			// スプライトグループ一覧の件数分、ループ
			for(int i = 0; i < countSpriteGroupLst; i++){

				// スプライトグループを取得
				PRSpriteGroup spriteGroup = spriteGroupLst.get(i);

				// スプライトグループがnull以外の場合、
				if(spriteGroup != null){

					// スプライトグループの時間を進める
					spriteGroup.forward();

					// スプライトグループを実行
					spriteGroup.run();

				}

			} // end for スプライトグループ一覧の件数分、ループ

		} // end if フェーズ−ゲーム中の場合

	} // end runGroup

	/**
	 * グループクリーン
	 * <pre>
	 * タイマーから定期的に呼ばれます。
	 * </pre>
	 */
	synchronized public void cleanGroup() {

		// フェーズ−ゲーム中の場合、
		if(phase == PHASE_GAME){

			// スプライトグループ一覧の件数を取得
			int countSpriteGroupLst = spriteGroupLst.size();

			// スプライトグループ一覧の件数が100件を超えた場合、
			if(countSpriteGroupLst > 100){

				// スプライトグループ一覧の件数分、ループ
				for(int i = 0; i < countSpriteGroupLst; i++){

					// スプライトグループを取得
					PRSpriteGroup spriteGroup = spriteGroupLst.get(i);

					// スプライトグループがnullの場合、
					if(spriteGroup == null){

						// スプライトグループをつめる
						spriteGroupLst.remove(i);

						// インデックス戻し
						i--;
						countSpriteGroupLst--;

					}

				} // end for スプライトグループ一覧の件数分、ループ

			} // end if スプライトグループ一覧の件数が100件を超えた場合

		} // end if フェーズ−ゲーム中の場合

	} // end cleanGroup

	/**
	 * 進行
	 * <pre>
	 * </pre>
	 */
	private void progress() {

		// ステージ１と２の場合、
		if(stage == 1 || stage == 2){

			// 低木を出現させる場合、
			if(Math.random() < 0.12){

				// 低木を生成
				PRSpriteShrub spriteShrub = new PRSpriteShrub(this, stage);

				// ノード追加
				this.addNode(spriteShrub);

			} // end if

		} // end if ステージ

	} // end progress

	/**
	 * フェーズ設定
	 */
	public void setPhase(int phase){

		// 時間初期化
		time = 0;

		// フェーズ変更
		this.phase = phase;

	} // end setPhase

	/**
	 * スコア加算
	 */
	public void addScore(int score){

		this.score = this.score + score;

	} // end addScore

	/**
	 * ゲーム中か取得
	 * @return boolean true−ゲーム中／ゲーム中ではない
	 */
	public boolean isPlaying(){
		return flgPlaying;
	}

	/**
	 * 衝突状態取得
	 */
	public boolean isBump(){
		return (phase == PHASE_BUMP);
	}

	/**
	 * ノード追加
	 * <pre>
	 * ノードを追加します。
	 * </pre>
	 * @param sprite ノード
	 */
	public void addNode(Node node){

		// 3Dパネルクラスを呼び出し
		panel3D.addNode(node);

	}

	/**
	 * ノード削除
	 * <pre>
	 * ノードを削除します。
	 * 当メソッドでは、ダミーノードに置き換え、runメソッドの中で削除します。
	 * nullに置き換えができない
	 * かつ Listクラスにremoveメソッドを多用することで配列コピーを多用されることを避けるため
	 * （重くなることをできるだけ避ける）
	 * </pre>
	 * @param sprite ノード
	 */
	public void removeNode(Node node){

		// 3Dパネルクラスを呼び出し
		panel3D.removeNode(node);

		// ロック対象であれば、外す
		lock.removeNode(node);

	}

	/**
	 * 影ノード追加
	 * <pre>
	 * ノードを追加します。
	 * </pre>
	 * @param sprite ノード
	 */
	public void addShadow(Node node){

		// 3Dパネルクラスを呼び出し
		panel3D.addShadow(node);

	}

	/**
	 * 影ノード追加
	 * <pre>
	 * 影ノードを追加します。
	 * </pre>
	 * @param sprite ノード
	 * @param radius 半径
	 */
	public void addShadow(Node node, double radius){

		// 3Dパネルクラスを呼び出し
		panel3D.addShadow(node, radius);

	}

	/**
	 * 影ノード削除
	 * <pre>
	 * 影ノードを削除します。
	 * </pre>
	 * @param sprite ノード
	 */
	public void removeShadow(Node node){

		// 3Dパネルクラスを呼び出し
		panel3D.removeShadow(node);

	}

	/**
	 * メディアビュー追加
	 * @param sprite スプライト
	 */
	public void addMediaView(PRSoundMP3 mp3){

		// スプライト一覧を取得
		ObservableList<Node> nodeLst = pane.getChildren();

		// メディアビューを追加
		nodeLst.add(mp3);

	} // end addMediaView

	/**
	 * メディアビュー削除
	 * @param sprite スプライト
	 */
	public void removeMediaView(PRSoundMP3 mp3){

		// スプライト一覧を取得
		ObservableList<Node> nodeLst = pane.getChildren();

		// インデックスを求める
		int index = nodeLst.indexOf(mp3);

		// インデックスが取得できた場合、
		if(index != -1){

			// 削除
			nodeLst.remove(mp3);

		}

	} // end removeMediaView

	/**
	 * スプライト一覧を取得
	 * @return ObservableList<Node> スプライト一覧
	 */
	public ObservableList<Node> getNodeLst(){

		// スプライト一覧を取得
		ObservableList<Node> nodeLst = panel3D.getChildren();

		// リターン
		return nodeLst;

	} // end getNodeLst

	/**
	 * スプライトグループ追加
	 * @param spriteGroup スプライト
	 */
	public void addSpriteGroup(PRSpriteGroup spriteGroup){

		// スプライトグループを追加
		spriteGroupLst.add(spriteGroup);

	} // end addSpriteGroup

	/**
	 * スプライトグループ削除
	 * @param spriteGroup スプライト
	 */
	public void removeSpriteGroup(PRSpriteGroup spriteGroup){

		// インデックスを求める
		int index = spriteGroupLst.indexOf(spriteGroup);

		// インデックスが取得できた場合、
		if(index != -1){

			// スプライトグループを削除
			spriteGroupLst.set(index, null);

		}

	} // end addSpriteGroup

	/**
	 * プレイヤー取得
	 */
	public PRSpritePrusher getPRSpritePrusher(){
		return spritePrusher;
	}

	/**
	 * ストーリー取得
	 */
	public PRStoryPrusher getPRStoryPrusher(){
		return storyPrusher;
	}

	/**
	 * ロック取得
	 */
	public PRLock getPRLock(){
		return lock;
	}

	/**
	 * ステージ取得
	 */
	public int getStage(){
		return stage;
	}

	/**
	 * 雲を開始
	 */
	public void startCloud(){

		// スプライトグループ雲を生成
		spriteGroupCloud = new PRSpriteGroupCloud(this);

		// アプリケーションにスプライトグループを追加
		this.addSpriteGroup(spriteGroupCloud);

	} // end startCloud

	/**
	 * 雲を終了させる
	 */
	public void finishCloud(){

		// スプライトグループ雲が存在する場合、
		if(spriteGroupCloud != null){

			// スプライトグループ雲を終了させる
			spriteGroupCloud.finish();

			// スプライトグループ雲を初期化する（オブジェクトを開放）
			spriteGroupCloud = null;

		} // スプライトグループ雲が存在する場合

	} // end finishCloud

	/**
	 * BGMスタート
	 */
	public void startMP3BGM(double millis){

		// ボリュームをデフォルト値に戻す
		mp3BGM.setVolume(1.0);

		// BGMをフェードアウトではない状態に戻す
		flgFadeout = false;

		// BGMをスタート
		mp3BGM.start(millis);

	} // end startMP3BGM

	/**
	 * BGMフェードアウト
	 * <pre>
	 * 時間の指定がない場合は、50フレーム⇒0.83秒でフェードアウトします。
	 * <pre>
	 */
	public void fadeoutMP3BGM(){

		// フェードアウトボリューム
		volumeFadeout = 0.02;

		// BGMをフェードアウト
		flgFadeout = true;

	} // end fadeoutMP3BGM

	/**
	 * BGMフェードアウト
	 * <pre>
	 * フレーム数で設定してください。
	 * 60フレームで設定すると1.00秒でフェードアウトします。
	 * <pre>
	 * @param frame フェードアウトするフレーム数。１秒＝60フレーム
	 */
	public void fadeoutMP3BGM(int frame){

		// フェードアウトボリューム
		volumeFadeout = 1.0 / frame;

		// BGMをフェードアウト
		flgFadeout = true;

	} // end fadeoutMP3BGM

	/**
	 * 天狗BGMスタート
	 */
	public void startMP3BGMTengu(){

		// BGMをスタート
		mp3BGMTengu.start(0);

	} // end startMP3BGMTengu

	/**
	 * 天狗BGMストップ
	 */
	public void stopMP3BGMTengu(){

		// BGMをストップ
		mp3BGMTengu.stop();

	} // end stopMP3BGMTengu

	/**
	 * レッド・アイBGMスタート
	 */
	public void startMP3BGMRedEye(){

		// BGMをスタート
		mp3BGMRedEye.start(0);

	} // end startMP3BGMRedEye

	/**
	 * レッド・アイBGMストップ
	 */
	public void stopMP3BGMRedEye(){

		// BGMをストップ
		mp3BGMRedEye.stop();

	} // end stopMP3BGMRedEye

	/**
	 * ファンファーレBGMスタート
	 */
	public void startMP3BGMFanfare(){

		// ファンファーレをスタート
		mp3BGMFanfare.start(0);

	} // end startMP3BGMFanfare

	/**
	 * ファンファーレBGMフェードアウト
	 */
	public void fadeoutMP3BGMFanfare(){

		// ファンファーレBGMをフェードアウト
		flgFadeoutFanfare = true;

	} // end fadeoutMP3FanfareBGM

	/**
	 * エンディングBGMスタート
	 */
	public void startMP3BGMEnding(){

		// エンディングをスタート
		mp3Ending.start();

	} // end startMP3BGMEnding

	/**
	 * ショット音スタート
	 */
	public void startWaveShot(){

		// 効果音を進める
		indexShot = (indexShot == 9)?0: ++indexShot;
		// 効果音スタート
		waveShots[indexShot].start();

	} // end startWaveShot

	/**
	 * 爆発音スタート
	 */
	public void startWaveBomb(){

		// 効果音を進める
		indexBomb = (indexBomb == 9)?0: ++indexBomb;
		// 効果音スタート
		waveBombs[indexBomb].start();

	} // end startWaveBomb

	/**
	 * 切り倒し音スタート
	 */
	public void startWaveHew(){

		// 効果音を進める
		indexHew = (indexHew == 9)?0: ++indexHew;
		// 効果音スタート
		waveHews[indexHew].start();

	} // end startWaveHew

	/**
	 * 跳ね返り音スタート
	 */
	public void startWaveRapel(){

		// 効果音を進める
		indexRapel = (indexRapel == 9)?0: ++indexRapel;
		// 効果音スタート
		waveRapels[indexRapel].start();

	} // end startWaveRapel

	/**
	 * 敵の弾音スタート
	 */
	public void startWaveTamaTeki(){

		// 効果音を進める
		indexTamaTeki = (indexTamaTeki == 9)?0: ++indexTamaTeki;
		// 効果音スタート
		waveTamaTekis[indexTamaTeki].start();

	} // end startWaveTamaTeki

	/**
	 * 敵のレーザー音１スタート
	 */
	public void startWaveLaser1Teki(){

		// 効果音を進める
		indexLaser1Teki = (indexLaser1Teki == 9)?0: ++indexLaser1Teki;
		// 効果音スタート
		waveLaser1Tekis[indexLaser1Teki].start();

	} // end startWaveLaser1Teki

	/**
	 * ロック音スタート
	 */
	public void startWaveLockon(){

		// 効果音を進める
		indexLockon = (indexLockon == 9)?0: ++indexLockon;
		// 効果音スタート
		waveLockons[indexLockon].start();

	} // end startWaveLockon

	/**
	 * 衝突
	 */
	public void bump(){

		// 効果音をスタート
		mp3Bump.start();

		// 時間初期化
		time = 0;

		// フェーズ−衝突へ
		phase = PHASE_BUMP;

	} // end bump

	/**
	 * カメラ設定
	 */
	private void initCamera(){

		// カメラ設定
		camera  = new PerspectiveCamera(true);

		// カメラの空間を設定
		camera.setFarClip(300);

		// 座標を設定
		camera.getTransforms().add(new Translate(0.0 , -1.0 , 0.0));

		// カメラを設定
		scene3D.setCamera(camera);

	} // end initCamera

	/**
	 * 背景移動X
	 * @param mx X移動量
	 */
	public void moveSkyX(double mx){
		spriteSky.moveX(mx);
	}

	/**
	 * カメラ移動Y
	 * @param my Y移動量
	 */
	public void moveCameraY(double my){
		camera.setTranslateY(camera.getTranslateY() + my);
	}

	/**
	 * 光源設定
	 */
	public void initLight(){

		// 光源設定
		if(light == null){

			light = new PointLight();

			// 光源を設定
			panel3D.getChildren().add(light);

		}

		// 光源位置
		light.setTranslateX(-100);
		light.setTranslateY(-200);
		light.setTranslateZ(-100);

	} // end initLight

	/**
	 * 光源設定
	 */
	public void initLight0(){

		// 光源位置
		light.setTranslateX(10);
		light.setTranslateY(-10);
		light.setTranslateZ(-100);

	} // end initLight

	/**
	 * スプライトグループ初期化
	 * <pre>
	 * スプライトグループを初期化します。
	 * 元々登録されていた場合には、クリアされます。
	 * </pre>
	 */
	public void initSpriteGroup(){

		// スプライトグループ一覧生成
		this.spriteGroupLst = new ArrayList<PRSpriteGroup>(100);

	}

	/**
	 * 敵スプライトをクリア
	 */
	public void removeSpriteTs(){

		// スプライト一覧を取得
		ObservableList<Node> nodeLst = this.getNodeLst();

		// スプライト一覧の件数を取得
		int countNodeLst = nodeLst.size();

		// スプライト一覧の件数分、ループ
		for(int i = 0; i < countNodeLst; i++){

			// スプライトを取得
			Node node = nodeLst.get(i);

			// 敵スプライトの場合、
			if(node instanceof PRSpriteT){
				// 敵スプライトを削除
				this.removeNode(node);
			}

		} // end for スプライト一覧の件数分、ループ

	} // end removeSpriteTs

	/**
	 * キー押下中かをどうかをチェック
	 * @param code キーコード
	 * @return boolean true：押している ／ false：押していない
	 */
	public boolean isKeyPressed(KeyCode code){

		return mapKeyPress.containsKey(code);

	} // end isKeyPressed

	/**
	 * キーを押したイベント
	 * <pre>
	 * ラムダ式記述を行って、インターフェース作成を省略しています。
	 * </pre>
	 */
	public void sceneKeyPressed(Event e){

		KeyEvent ke = (KeyEvent)e;
		KeyCode code = ke.getCode();
		mapKeyPress.put(code, "");

	} // end sceneKeyPressed

	/**
	 * キーを離したイベント
	 * <pre>
	 * ラムダ式記述を行って、インターフェース作成を省略しています。
	 * </pre>
	 */
	public void sceneKeyReleased(Event e){

		KeyEvent ke = (KeyEvent)e;
		KeyCode code = ke.getCode();
		mapKeyPress.remove(code);

	} // end sceneKeyReleased

	/**
	 * ウインドウサイズ変更イベント
	 */
	public void windowResized(){

		// ペイン（ウインドウ）のサイズを取得
		double widthPane = PRApplication.this.pane.getWidth();
		double heightPane = PRApplication.this.pane.getHeight();

		// 比率を求める
		double ratioWidth = (float)widthPane / PRConstant.WIDTH;
		double ratioHeight = (float)heightPane / PRConstant.HEIGHT;

		// 比率の小さい方を求める
		double ratio = (ratioWidth < ratioHeight)?ratioWidth: ratioHeight;

		// 各シーンを拡大縮小する
		scene3D.setScaleX(ratio);
		scene3D.setScaleY(ratio);
		scene2D.setScaleX(ratio);
		scene2D.setScaleY(ratio);

		// ペイン上のシーンを再配置する
		scene3D.setLayoutX((widthPane - PRConstant.WIDTH) / 2);
		scene3D.setLayoutY((heightPane - PRConstant.HEIGHT) / 2);
		scene2D.setLayoutX((widthPane - PRConstant.WIDTH) / 2);
		scene2D.setLayoutY((heightPane - PRConstant.HEIGHT) / 2);

	} // end windowResized

}
