import javafx.scene.Group;
import javafx.scene.image.Image;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.MeshView;
import javafx.scene.shape.TriangleMesh;
import javafx.scene.transform.Rotate;

/**
 * エンディング
 */
public class PRSpriteEnding extends PRSprite implements PRConstant {

	/** スプライト定義 */
	private static final String	ENDING_CONGRATULATIONS	= "PRSpriteEnding_Congratulations";
	private static final String ENDING_THEEND			= "PRSpriteEnding_TheEnd";

	private static float[]		POINTS_2		= {
		-48.0f, -360.0f, 0,
		-48.0f,  360.0f, 0,
		 48.0f,  360.0f, 0,
		 48.0f, -360.0f, 0
	};

	private static float[]		TEXCOORDS  		 = {
		0, 0,
		0, 1,
		1, 1,
		1, 0
	};

	private static int[]		FACES			= {
		0, 0, 1, 1, 2, 2,
		2, 2, 3, 3, 0, 0
	};

	/** イメージ */
	private static Image		image			= null;

	/** グループ */
	private		Group			group			= null;

	/** メッシュビュー */
	private		MeshView		meshView		= null;

	/** 時間 */
	private 	int				time			= 0;

	/**
	 * 初期化
	 */
	public static void init() {

		try{

			ObjReader.init(ENDING_CONGRATULATIONS , "3d/Congratulations.obj");
			ObjReader.init(ENDING_THEEND, "3d/TheEnd.obj");

			PRSpriteEnding.image = PRImage.getImage("3d/image_Ending01.png");

		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end init

	/**
	 * コンストラクタ
	 * @param application PRApplication
	 */
	public PRSpriteEnding(PRApplication application){

		super(application);

	} // end PRSpriteEnding

	/**
	 * 実行
	 */
	@Override
	public void run(){

		// 時間＋１
		time++;

		// はじめの時間
		if(time == 1){

			// 光源設定
			application.initLight0();

			// スプライトを生成
			group = new Group();

			// グループを登録
			this.getChildren().add(group);

			// オブジェクト
			ObjReader.create(group, ENDING_CONGRATULATIONS);

			// 移動
			group.setTranslateY(-1.0);

		// 6秒
		}else if(time < 360){

			// 移動
			group.setTranslateZ(time / 360.0 * 15);

		// 12秒
		}else if(time < 720){

			// 何もしない

		// 19秒
		}else if(time < 1160){

			// 移動
			group.setTranslateZ(15 + (time - 720.0) / (1160 - 720) * 300);

			// 12秒
			if(time == 720){

				// ファンファーレBGMフェードアウト
				application.fadeoutMP3BGMFanfare();

			}

		// 19秒
		}else if(time == 1160){

			// グループを削除
			this.getChildren().remove(group);

			// 光源設定
			application.initLight();

			// エンディングBGMスタート
			application.startMP3BGMEnding();

			// メッシュビューを生成
			meshView = new MeshView();

			// メッシュを生成
			TriangleMesh mesh = new TriangleMesh();

			mesh.getPoints().addAll(POINTS_2);
			mesh.getTexCoords().addAll(TEXCOORDS);
			mesh.getFaces().addAll(FACES);

			// マテリアルを生成
			PhongMaterial material = new PhongMaterial();
			material.setDiffuseMap(image);

			// メッシュを登録
			meshView.setMesh(mesh);
			meshView.setMaterial(material);

			// 回転
			meshView.getTransforms().add(new Rotate(-72.0, Rotate.X_AXIS));

			// 位置を修正
			meshView.setTranslateY(144.0);
			meshView.setTranslateZ(-233.0);

			// メッシュビューを追加
			this.getChildren().add(meshView);

		// 69.5秒
		}else if(time < 4170){

			// 移動
			meshView.setTranslateY(meshView.getTranslateY() - 0.3090 / 3.3);
			meshView.setTranslateZ(meshView.getTranslateZ() + 0.9511 / 3.3);

		// 70秒
		}else if(time == 4200){

			// メッシュを削除
			this.getChildren().remove(group);

			// 光源設定
			application.initLight0();

			// スプライトを生成
			group = new Group();

			// グループを登録
			this.getChildren().add(group);

			// オブジェクト
			ObjReader.create(group, ENDING_THEEND);

			// 移動
			group.setTranslateY(-1.0);

		// 73秒
		}else if(time < 4380){

			// 移動
			group.setTranslateZ(6 + (4380.0 - time) / (4380.0 - 4200) * 300);

		// 83秒
		}else if(time < 4980){

			// 何もしない

		// 88秒
		}else if(time < 5280){

			// 移動
			group.setTranslateZ(6 + (time - 4980) / (5280.0 - 5160) * 300);

		// 90秒
		}else if(time < 5400){

			// 何もしない

		// 90秒
		}else if(time == 5400){

			// グループを削除
			this.getChildren().remove(group);

			// エンディングを削除（自分自身を削除）
			this.remove();

			// 光源設定
			application.initLight();

			// フェーズ設定
			super.application.setPhase(PHASE_INIT);

		} // end if 時間

	} // end run

}
