import javafx.scene.image.Image;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.MeshView;
import javafx.scene.shape.TriangleMesh;
import javafx.scene.paint.Color;
import javafx.collections.ObservableList;
import javafx.scene.Node;

/**
 * 空
 */
public class PRSpriteSky extends PRSprite {

	private static float[]		POINTS			= {
		-100.0f, -100, 0,
		-100.0f,  100, 0,
		 100.0f,  100, 0,
		 100.0f, -100, 0
	};

	private static float[]		TEXCOORDS  		 = {
		0, 0,
		0, 1,
		1, 1,
		1, 0
	};

	private static int[]		FACES			= {
		0, 0, 1, 1, 2, 2,
		2, 2, 3, 3, 0, 0
	};

	private static float[]		POINTS_SKY1		= {
		-100.0f, -112, 0,
		-100.0f,  -32, 0,
		 100.0f,  -32, 0,
		 100.0f, -112, 0
	};

	private static float[]		POINTS_SKY2		= {
		-100.0f,  -30, 0,
		-100.0f,   00, 0,
		 100.0f,   00, 0,
		 100.0f,  -30, 0
	};

	/** イメージ */
	private static Image[]		images			= null;

	/** マテリアル */
	private		PhongMaterial	material		= null;

	/** 空１ノード一覧 */
	private		MeshView[]		meshViewSky1s	= null;

	/** 空２ノード一覧 */
	private		MeshView[]		meshViewSky2s	= null;

	/**
	 * 初期化
	 */
	public static void init() {

		try{
			PRSpriteSky.images = new Image[]{
				PRImage.getImage("3d/imageStage1-1.png"),
				PRImage.getImage("3d/imageStage1-2.png"),
				PRImage.getImage("3d/imageStage2-1.png"),
				PRImage.getImage("3d/imageStage2-2.png"),
			};
		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end init

	/**
	 * コンストラクタ
	 * @param application PRApplication
	 */
	public PRSpriteSky(PRApplication application){

		// スーパークラスを呼び出す
		super(application);

		try{

			// Y設定
			super.setTranslateY(1.0);

			// メッシュビューを生成
			MeshView meshView = new MeshView();

			// メッシュを生成
			TriangleMesh mesh = new TriangleMesh();

			mesh.getPoints().addAll(POINTS);
			mesh.getTexCoords().addAll(TEXCOORDS);
			mesh.getFaces().addAll(FACES);

			// マテリアルを生成
			material = new PhongMaterial();
			material.setDiffuseColor(Color.DARKTURQUOISE);

			// 位置を修正
			meshView.setTranslateZ(300); 

			// メッシュを登録
			meshView.setMesh(mesh);
			meshView.setMaterial(material);

			// メッシュビューを追加
			this.getChildren().add(meshView);

			// ステージ１で設定
			this.setStage(1);

		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end PRSpriteSky

	/**
	 * ステージ設定
	 * @param stage ステージ
	 */
	public void setStage(int stage){

		// ノード一覧を取得
		ObservableList<Node> nodeLst = this.getChildren();

		// ノード一覧の件数を取得
		int countNodeLst = nodeLst.size();

		// ノード一覧からノードを削除
		// ０（空）は含まない
		nodeLst.remove(1, countNodeLst);

		// ステージ１の場合、
		if(stage == 1){

			// 空の色をダーク・ターコイズ色へ
			material.setDiffuseColor(Color.DARKTURQUOISE);

			// 空１ノード一覧を生成
			meshViewSky1s = new MeshView[2];

			// ２件、ループ
			for(int i = 0; i < 2; i++){

				// メッシュビューを生成
				MeshView meshView = new MeshView();

				// メッシュを生成
				TriangleMesh mesh = new TriangleMesh();

				mesh.getPoints().addAll(POINTS_SKY1);
				mesh.getTexCoords().addAll(TEXCOORDS);
				mesh.getFaces().addAll(FACES);

				// マテリアルを生成
				PhongMaterial material = new PhongMaterial();
				material.setDiffuseMap(images[0]);

				// メッシュを登録
				meshView.setMesh(mesh);
				meshView.setMaterial(material);

				// 位置を修正
				meshView.setTranslateX(i * 200);
				meshView.setTranslateZ(299.9999); 

				// メッシュビューを追加
				this.getChildren().add(meshView);

				// 空１ノード一覧に設定
				meshViewSky1s[i] = meshView;

			} // end for １件、ループ

			// 空２ノード一覧を生成
			meshViewSky2s = new MeshView[2];

			// ２件、ループ
			for(int i = 0; i < 2; i++){

				// メッシュビューを生成
				MeshView meshView = new MeshView();

				// メッシュを生成
				TriangleMesh mesh = new TriangleMesh();

				mesh.getPoints().addAll(POINTS_SKY2);
				mesh.getTexCoords().addAll(TEXCOORDS);
				mesh.getFaces().addAll(FACES);

				// マテリアルを生成
				PhongMaterial material = new PhongMaterial();
				material.setDiffuseMap(images[1]);

				// メッシュを登録
				meshView.setMesh(mesh);
				meshView.setMaterial(material);

				// 位置を修正
				meshView.setTranslateX(i * 200);
				meshView.setTranslateZ(299.9999); 

				// メッシュビューを追加
				this.getChildren().add(meshView);

				// 空２ノード一覧に設定
				meshViewSky2s[i] = meshView;

			} // end for ２件、ループ

		// ステージ２の場合、
		}else if(stage == 2){

			// 空の色をカーキ色へ
			material.setDiffuseColor(Color.KHAKI);

			// 空１ノード一覧を生成
			meshViewSky1s = new MeshView[2];

			// ２件、ループ
			for(int i = 0; i < 2; i++){

				// メッシュビューを生成
				MeshView meshView = new MeshView();

				// メッシュを生成
				TriangleMesh mesh = new TriangleMesh();

				mesh.getPoints().addAll(POINTS_SKY1);
				mesh.getTexCoords().addAll(TEXCOORDS);
				mesh.getFaces().addAll(FACES);

				// マテリアルを生成
				PhongMaterial material = new PhongMaterial();
				material.setDiffuseMap(images[2]);

				// メッシュを登録
				meshView.setMesh(mesh);
				meshView.setMaterial(material);

				// 位置を修正
				meshView.setTranslateX(i * 200);
				meshView.setTranslateZ(299.9999); 

				// メッシュビューを追加
				this.getChildren().add(meshView);

				// 空１ノード一覧に設定
				meshViewSky1s[i] = meshView;

			} // end for １件、ループ

			// 空２ノード一覧を生成
			meshViewSky2s = new MeshView[2];

			// ２件、ループ
			for(int i = 0; i < 2; i++){

				// メッシュビューを生成
				MeshView meshView = new MeshView();

				// メッシュを生成
				TriangleMesh mesh = new TriangleMesh();

				mesh.getPoints().addAll(POINTS_SKY2);
				mesh.getTexCoords().addAll(TEXCOORDS);
				mesh.getFaces().addAll(FACES);

				// マテリアルを生成
				PhongMaterial material = new PhongMaterial();
				material.setDiffuseMap(images[3]);

				// メッシュを登録
				meshView.setMesh(mesh);
				meshView.setMaterial(material);

				// 位置を修正
				meshView.setTranslateX(i * 200);
				meshView.setTranslateZ(299.9999); 

				// メッシュビューを追加
				this.getChildren().add(meshView);

				// 空２ノード一覧に設定
				meshViewSky2s[i] = meshView;

			} // end for ２件、ループ

		// ステージ３の場合、
		}else if(stage == 3){

			// 空の色をドジャー・ブルー色へ
			material.setDiffuseColor(Color.DODGERBLUE);

		} // end if ステージ

	} // end setStage

	/**
	 * X移動
	 * @param mx X移動量
	 */
	public void moveX(double mx){

		// 画面戻し制御
		int ctrlModosi = 0;

		// 空１ノード一覧の件数を取得
		int countMeshViewsSky1 = meshViewSky1s.length;

		// 空１ノード一覧の件数分、ループ
		for(int i = 0; i < countMeshViewsSky1; i++){

			// 空１ノードを取得
			MeshView meshView = meshViewSky1s[i];

			// 座標取得
			double x = meshView.getTranslateX();

			// 座標加算
			x = x + mx / 4;

			// ０件目の場合、
			if(i == 0){

				// Xが-200を超えた場合、
				if(x < -200){
					// 右に戻す
					ctrlModosi = 1;

				// Xが0を超えた場合、
				}else if(x > 0){
					// 画面戻しフラグをON
					ctrlModosi = 2;
				}

			} // end for ０件目の場合

			// 右に戻す場合、
			if(ctrlModosi == 1){

				// 右に戻す
				x = x + 200;

			// 左に戻す場合、
			}else if(ctrlModosi == 2){

				// 左に戻す
				x = x - 200;

			}

			// 移動
			meshView.setTranslateX(x);

		} // end for 空１ノード一覧の件数分、ループ

		// 空２ノード一覧の件数を取得
		int countMeshViewsSky2 = meshViewSky2s.length;

		// 空２ノード一覧の件数分、ループ
		for(int i = 0; i < countMeshViewsSky2; i++){

			// 空２ノードを取得
			MeshView meshView = meshViewSky2s[i];

			// 座標取得
			double x = meshView.getTranslateX();

			// 座標加算
			x = x + mx;

			// ０件目の場合、
			if(i == 0){

				// Xが-200を超えた場合、
				if(x < -200){
					// 右に戻す
					ctrlModosi = 1;

				// Xが0を超えた場合、
				}else if(x > 0){
					// 画面戻しフラグをON
					ctrlModosi = 2;
				}

			} // end for ０件目の場合

			// 右に戻す場合、
			if(ctrlModosi == 1){

				// 右に戻す
				x = x + 200;

			// 左に戻す場合、
			}else if(ctrlModosi == 2){

				// 左に戻す
				x = x - 200;

			}

			// 移動
			meshView.setTranslateX(x);

		} // end for 空２ノード一覧の件数分、ループ

	} // end moveX

	/**
	 * 実行
	 */
	@Override
	public void run(){

	} // end run

}
