import javafx.collections.ObservableList;
import javafx.geometry.BoundingBox;
import javafx.geometry.Bounds;
import javafx.scene.Node;
import javafx.scene.image.Image;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.MeshView;
import javafx.scene.shape.TriangleMesh;

/**
 * 石
 */
public class PRSpriteStone extends PRSpriteT {

	private static float[]		POINTS		= {
		-1,-0.5f, 0,
		-1, 0.5f, 0,
		 1, 0.5f, 0,
		 1,-0.5f, 0
	};

	private static float[]		TEXCOORDS   = {
		0, 0,
		0, 1,
		1, 1,
		1, 0
	};

	private static int[]		FACES		= {
		0, 0, 1, 1, 2, 2,
		2, 2, 3, 3, 0, 0
	};

	/** イメージ */
	private static Image		image		= null;

	/**
	 * 初期化
	 */
	public static void init() {

		try{
			PRSpriteStone.image = PRImage.getImage("3d/image_publicdomainq-0013646jjvkqj.png");
		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end init

	/**
	 * コンストラクタ
	 * @param application PRApplication
	 * @param stage ステージ
	 */
	public PRSpriteStone(PRApplication application){

		// スーパークラスを呼び出す
		super(application);

		try{

			// メッシュビューを生成
			MeshView meshView = new MeshView();

			// メッシュを生成
			TriangleMesh mesh = new TriangleMesh();

			mesh.getPoints().addAll(POINTS);
			mesh.getTexCoords().addAll(TEXCOORDS);
			mesh.getFaces().addAll(FACES);

			// マテリアルを生成
			PhongMaterial material = new PhongMaterial();
			material.setDiffuseMap(image);

			// メッシュ／マテリアルを登録
			meshView.setMesh(mesh);
			meshView.setMaterial(material);

			// メッシュビューを追加
			this.getChildren().add(meshView);

			// 拡大／縮小する
			super.setScaleX(0.5);
			super.setScaleY(0.5);
			super.setScaleZ(0.5);

			// X設定
			double x = Math.random() * 24.0 - 12.0;
			super.setTranslateX(x);

			// Y設定
			Bounds bounds = this.getBoundsInParent();
			super.setTranslateY(-bounds.getHeight() / 2);

			// Z設定
			super.setTranslateZ(100);

			// 速度設定
			if(application.getStage() == 1){
				mz = -0.6;
			}else{
				mz = -1.0;
			}

			// 影を設定
			application.addShadow(this, 0.6);

		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end PRSpriteStone

	/**
	 * 実行
	 */
	@Override
	public void run(){

		//
		// X方向
		//

		// 移動量取得
		double mx = -application.getPRSpritePrusher().mx;

		// 座標計算
		double x = super.getTranslateX();
		x = x + mx;

		// 移動
		super.setTranslateX(x);

		//
		// Z方向
		//

		// 座標取得
		double z = super.getTranslateZ();

		// 座標加算
		z = z + mz;

		// 画面の外にでた場合、
		if(z < 0){
			// 削除する
			super.remove();
		}

		// 移動
		super.setTranslateZ(z);

		//
		// 当たり判定
		//

		// 爆発していない場合、
		if(!flgBomb){

			// 境界を取得
			Bounds boundsThis = this.getBoundsInParent();

			// すり抜ける時があるので、奥行きを持たせる
			// Boundsは、protectedで生成できないので、BoundingBoxで生成する
			boundsThis = new BoundingBox(
				boundsThis.getMinX(),
				boundsThis.getMinY(),
				boundsThis.getMinZ(),
				boundsThis.getWidth(),
				boundsThis.getHeight(),
				boundsThis.getDepth() + 2.5
			);

			// スプライト一覧を取得
			ObservableList<Node> nodeLst = application.getNodeLst();

			// スプライト一覧の件数を取得
			int countNodeLst = nodeLst.size();

			// スプライト一覧の件数分、ループ
			for(int i = 0; i < countNodeLst; i++){

				// スプライトを取得
				Node node = nodeLst.get(i);

				// 弾の場合、
				if(node instanceof PRSpriteTama){

					// 当たっている場合、
					if(boundsThis.intersects(node.getBoundsInParent())){

						// 爆発
						flgBomb = true;

						// 弾を削除
						application.removeNode(node);

						// 爆発を生成
						PRSpriteBomb spriteBomb = new PRSpriteBomb(application, this, 0.04);

						// アプリケーションに爆発を追加
						application.addNode(spriteBomb);

						// 点数追加
						application.addScore(100);

						// 爆発音スタート
						application.startWaveBomb();

						// ループを抜ける
						break;

					} // end if 当たっている場合

				} // end if 弾の場合

			} // end for スプライト一覧の件数分、ループ

		} // end if 爆発していない場合

	} // end run

	/**
	 * 当たり判定範囲X
	 */
	@Override
	public double getHaniX() {
		return 0.1;
	}

	@Override
	public double getHaniY() {
		return 0.8;
	}

}
