import javafx.geometry.Point3D;
import javafx.collections.ObservableList;
import javafx.scene.Node;
import javafx.geometry.Bounds;

/**
 * 天狗
 */
public class PRSpriteTengu extends PRSpriteT {

	/** 配置パターン */
	private		String					strPattern		= null;

	/** 回転速度 */
	private		double					s				= Math.random() * 4 + 2;

	/** 敵の弾発射状態 */
	private		boolean					flgTamaTeki		= false;

	/** 角度 */
	private		int						angle			= 0;

	/** 敵の弾発射時間 */
	private		int						timeTamaTeki	= 240;

	/** スプライトグループ */
	private		PRSpriteGroupTengu		spriteGroup		= null;

	/**
	 * 初期化
	 */
	public static void init() {

		try{
			ObjReader.init(PRSpriteTengu.class, "3d/Tengu.obj");
		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end init

	/**
	 * コンストラクタ
	 * @param application PRApplication
	 * @param strParam 配置パターン
	 */
	public PRSpriteTengu(PRApplication application, String strParam){

		// スーパークラスを呼び出す
		super(application);

		try{

			// パターンを保持
			this.strPattern = strParam;

			// 直進の場合、
			if("直進".equals(strPattern)){

				// 座標設定
				super.setTranslateX(Math.random() * 10.0 - 5.0);
				super.setTranslateY(Math.random() - 3.0);
				super.setTranslateZ(100);

				// 速度設定
				mz = -1.0;

				// オブジェクト
				ObjReader.create(this);

				// 回転軸
				Point3D p = new Point3D(0.0, 1.0, 0.0);

				// 回転軸を設定
				super.setRotationAxis(p);

			// 右からの攻撃の場合、
			}else if("右からの攻撃".equals(strPattern)){

				// 座標設定
				super.setTranslateX(4.5);
				super.setTranslateY(Math.random() - 4.5 + application.getPRSpritePrusher().getTranslateY());
				super.setTranslateZ(30);

				// 速度設定
				mx = -0.05;
				my =  0.06;
				mz = -0.40;

				// オブジェクト
				ObjReader.create(this);

				// 回転軸
				Point3D p = new Point3D(0.0, 1.0, 0.0);

				// 回転軸を設定
				super.setRotationAxis(p);

			// 左からの攻撃の場合、
			}else if("左からの攻撃".equals(strPattern)){

				// 座標設定
				super.setTranslateX(-4.5);
				super.setTranslateY(Math.random() - 4.5 + application.getPRSpritePrusher().getTranslateY());
				super.setTranslateZ(30);

				// 速度設定
				mx =  0.05;
				my =  0.06;
				mz = -0.40;

				// オブジェクト
				ObjReader.create(this);

				// 回転軸
				Point3D p = new Point3D(0.0, -1.0, 0.0);

				// 回転軸を設定
				super.setRotationAxis(p);

			} // end if パターン

			// 影を設定
			application.addShadow(this, 0.6);

		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end PRSpriteTengu

	/**
	 * コンストラクタ
	 * @param application PRApplication
	 * @param spriteGroup スプライトグループ
	 * @param strParam 配置パターン
	 * @param angle 角度
	 */
	public PRSpriteTengu(PRApplication application, PRSpriteGroupTengu spriteGroup, String strParam, int angle){

		// スーパークラスを呼び出す
		super(application);

		try{

			// スプライトグループを保持
			this.spriteGroup = spriteGroup;

			// 角度を保持
			this.angle = angle;

			// パターンを保持
			this.strPattern = strParam;

			// ボス１の場合、
			if("ボス１".equals(strPattern)){

				// 拡大する
				super.setScaleX(2.0);
				super.setScaleY(2.0);
				super.setScaleZ(2.0);

				// ラジアンを計算
				double rad = angle * Math.PI / 180;

				// 座標設定
				super.setTranslateX(Math.sin(rad) * 3.6);
				super.setTranslateY(Math.cos(rad) * 3.6 - 6.0);
				super.setTranslateZ(150);

				// 速度設定
				mz = -1.0;

				// オブジェクト
				ObjReader.create(this);

			} // end if パターン

		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end PRSpriteTengu

	/**
	 * 実行
	 */
	@Override
	public void run(){

		// 座標取得
		double x = super.getTranslateX();
		double y = super.getTranslateY();
		double z = super.getTranslateZ();

		// 直進の場合、
		if("直進".equals(strPattern)){

			// 座標加算
			z = z + mz;

			// 画面の外にでた場合、
			if(z < 0){
				// 削除する
				super.remove();
			}

			// 移動
			super.setTranslateZ(z);

			// 回転
			super.setRotate(getRotate() + this.s);

		// 右からの攻撃の場合、
		}else if("右からの攻撃".equals(strPattern)){

			// 座標加算
			x = x + mx;
			y = y + my;
			z = z + mz;

			// 方向転換１回目
			if(z < 8.0){
				z = 8.0;
				my = 0.0;
				mz = 0.0;
			}

			// 方向転換２回目
			if(x < -2.0){
				my = -0.06;
				mz =  0.40;
			} // end if 方向転換１回目

			// 敵の弾を発射していない場合、
			if(x < -4.0 && !flgTamaTeki){

				// 敵の弾発射
				flgTamaTeki = true;

				// 敵の弾を生成
				PRSpriteTamaTeki spriteTamaTeki = new PRSpriteTamaTeki(application, this, 0.2);
				// 敵の弾を追加
				application.addNode(spriteTamaTeki);

				// 敵の弾音スタート
				application.startWaveTamaTeki();

			} // end if 敵の弾を発射していない場合

			// 画面の外にでた場合、
			if(x < -5.0){
				// 削除する
				super.remove();
			}

			// 移動
			super.setTranslateX(x);
			super.setTranslateY(y);
			super.setTranslateZ(z);

			// 回転
			super.setRotate(getRotate() + this.s);

		// 左からの攻撃の場合、
		}else if("左からの攻撃".equals(strPattern)){

			// 座標加算
			x = x + mx;
			y = y + my;
			z = z + mz;

			// 方向転換１回目
			if(z < 8.0){
				z = 8.0;
				my = 0.0;
				mz = 0.0;
			}

			// 方向転換２回目
			if(x > 2.0){
				my = -0.06;
				mz =  0.40;
			} // end if 方向転換１回目

			// 敵の弾を発射していない場合、
			if(x > 4.0 && !flgTamaTeki){

				// 敵の弾発射
				flgTamaTeki = true;

				// 敵の弾を生成
				PRSpriteTamaTeki spriteTamaTeki = new PRSpriteTamaTeki(application, this, 0.2);
				// 敵の弾を追加
				application.addNode(spriteTamaTeki);

				// 敵の弾音スタート
				application.startWaveTamaTeki();

			} // end if 敵の弾を発射していない場合

			// 画面の外にでた場合、
			if(x > 5.0){
				// 削除する
				super.remove();
			}

			// 移動
			super.setTranslateX(x);
			super.setTranslateY(y);
			super.setTranslateZ(z);

			// 回転
			super.setRotate(getRotate() + this.s);

		// ボス１の場合、
		}else if("ボス１".equals(strPattern)){

			// 座標加算
			z = z + mz;

			// 停止位置にきた場合、
			if(z < 20){
				// 速度を０に設定
				mz = 0;
			}

			// 角度を加算
			angle = angle + 2;

			// ラジアンを計算
			double rad = angle * Math.PI / 180;

			// 移動
			super.setTranslateX(Math.sin(rad) * 3.6);
			super.setTranslateY(Math.cos(rad) * 3.6 - 6.0);
			super.setTranslateZ(z);

			// 敵の弾時間−１
			timeTamaTeki--;

			// 敵の弾発射時間が来たら
			if(timeTamaTeki <= 0){

				// 敵の弾を生成
				PRSpriteTamaTeki spriteTamaTeki = new PRSpriteTamaTeki(application, this, 0.30);
				// 敵の弾を追加
				application.addNode(spriteTamaTeki);

				// 敵の弾音スタート
				application.startWaveTamaTeki();

				// 敵の弾発射時間 */
				timeTamaTeki	= (int)(Math.random() * 60) + 180;

			} // end if 敵の弾を発射していない場合

		} // end if パターン

		//
		// 当たり判定
		//

		// 爆発していない場合、
		if(!flgBomb){

			// 境界を取得
			Bounds boundsThis = this.getBoundsInParent();

			// スプライト一覧を取得
			ObservableList<Node> nodeLst = application.getNodeLst();

			// スプライト一覧の件数を取得
			int countNodeLst = nodeLst.size();

			// スプライト一覧の件数分、ループ
			for(int i = 0; i < countNodeLst; i++){

				// スプライトを取得
				Node node = nodeLst.get(i);

				// 弾の場合、
				if(node instanceof PRSpriteTama){

					// 当たっている場合、
					if(boundsThis.intersects(node.getBoundsInParent())){

						// 爆発
						flgBomb = true;

						// 弾を削除
						application.removeNode(node);

						// 爆発を生成
						PRSpriteBomb spriteBomb = new PRSpriteBomb(application, this);

						// アプリケーションに爆発を追加
						application.addNode(spriteBomb);

						// 点数追加
						application.addScore(500);

						// 爆発音スタート
						application.startWaveBomb();

						// ボス１の場合、
						if("ボス１".equals(strPattern)){
							// スプライトグループに削除を通知
							spriteGroup.removeSprite();
						}

						// ループを抜ける
						break;

					} // end if 当たっている場合

				} // end if 弾の場合

			} // end for スプライト一覧の件数分、ループ

		} // end if 爆発していない場合

	} // end run

	/**
	 * 当たり判定範囲X
	 */
	@Override
	public double getHaniX() {
		return 0.4;
	}

	/**
	 * 当たり判定範囲Y
	 */
	@Override
	public double getHaniY() {
		return 0.4;
	}

	/**
	 * 当たり判定範囲Z
	 */
	@Override
	public double getHaniZ() {
		return 0.4;
	}

}
