import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.MeshView;
import javafx.scene.shape.TriangleMesh;

/**
 * タイル
 */
public class PRSpriteTile extends PRSprite {

	private static float[]		POINTS		= {
		-1.0f, 0,  2.0f,
		-1.0f, 0, -2.0f,
		 1.0f, 0, -2.0f,
		 1.0f, 0,  2.0f
	};

	private static float[]		TEXCOORDS   = {
		0, 0,
		0, 1,
		1, 1,
		1, 0
	};

	private static int[]		FACES		= {
		0, 0, 1, 0, 2, 0,
		2, 0, 3, 0, 0, 0
	};

	/** 速度Z */
	private		double			mz			= -0.6;

	/** マテリアル */
	private		PhongMaterial	material	= null;

	/**
	 * 初期化
	 */
	public static void init() {

		try{
		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end init

	/**
	 * コンストラクタ
	 * @param application PRApplication
	 */
	public PRSpriteTile(PRApplication application){

		// スーパークラスを呼び出す
		super(application);

		try{

			// 位置補正
			int posX = 0;

			// マテリアルを生成
			material = new PhongMaterial();
			material.setDiffuseColor(Color.FORESTGREEN);

			// zループ
			for(int z = 200; z > 0; z = z - 4){

				// 位置補正
				posX = (posX == 0)?1: 0;

				// xループ
				for(int x = -32; x < 32; x = x + 4){

					// メッシュビューを生成
					MeshView meshView = new MeshView();

					// メッシュを生成
					TriangleMesh mesh = new TriangleMesh();

					mesh.getPoints().addAll(POINTS);
					mesh.getTexCoords().addAll(TEXCOORDS);
					mesh.getFaces().addAll(FACES);

					// メッシュを登録
					meshView.setMesh(mesh);
					meshView.setMaterial(material);

					// X設定
					meshView.setTranslateX(x + posX * 2);

					// Y設定
					meshView.setTranslateY(-0.0001);

					// Z設定
					meshView.setTranslateZ(z);

					// メッシュビューを追加
					this.getChildren().add(meshView);

				} // end for xループ

			} // end for zループ

		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end PRSpriteTile

	/**
	 * ステージ設定
	 * @param stage ステージ
	 */
	public void setStage(int stage){

		// ステージ１：フォレスト・グリーン
		if(stage == 1){

			this.setVisible(true);
			material.setDiffuseColor(Color.FORESTGREEN);

			mz = -0.6;

		// ステージ２：ゴールデンロッド・スピードアップ
		}else if(stage == 2){

			material.setDiffuseColor(Color.GOLDENROD);

			mz = -1.0;

		// ステージ３：隠す
		}else if(stage == 3){

			this.setVisible(false);

		} // end if ステージ

	} // end setStage

	/**
	 * 実行
	 */
	@Override
	public void run(){

		//
		// X方向
		//

		// 座標取得
		double x = super.getTranslateX();

		// 移動量取得
		double mx = -application.getPRSpritePrusher().mx;

		// 座標計算
		x = x + mx;

		// 座標再配置
		if(x < -10){
			x = x + 10;
		}else if(x > 10){
			x = x - 10;
		}

		// 移動
		super.setTranslateX(x);

		//
		// Z方向
		//

		// 座標取得
		double z = super.getTranslateZ();

		// 座標加算
		z = z + mz;

		// ループタイミングの場合、
		if(z <= -16){
			// 位置を戻す
			z = 0;;
		}

		// 移動
		super.setTranslateZ(z);

	} // end run

}
