import java.util.HashMap;

import javafx.geometry.Point3D;
import javafx.scene.Group;
import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.MeshView;
import javafx.scene.transform.Rotate;
import javafx.scene.transform.Translate;

/**
 * タイトル−Prusher
 */
public class PRSpriteTitlePrusher extends PRSprite implements PRConstant {

	/** スプライト定義 */
	private static final String	TITLE_P			= "PRSpriteTITLE_Prusher_P";
	private static final String	TITLE_R1		= "PRSpriteTITLE_Prusher_R";
	private static final String	TITLE_U			= "PRSpriteTITLE_Prusher_U";
	private static final String	TITLE_S			= "PRSpriteTITLE_Prusher_S";
	private static final String	TITLE_H			= "PRSpriteTITLE_Prusher_H";
	private static final String	TITLE_E			= "PRSpriteTITLE_Prusher_E";
	private static final String	TITLE_R2		= "PRSpriteTITLE_Prusher_R";

	private static final String[]	TitleS		= {
		TITLE_P	,
		TITLE_R1,
		TITLE_U	,
		TITLE_S	,
		TITLE_H	,
		TITLE_E	,
		TITLE_R2
	};

	/** 初期設定用 */
	private static final Rotate INIT_ROTATE		= new Rotate(90, Rotate.X_AXIS);
	private static final Translate INIT_TRANSLATE = new Translate(0.0, 0.25, -0.25);

	/** スプライト件数 */
	private static final int	COUNT_SPRITES	= 7;

	/** Z固定位置 */
	private static final double	Z				= 8.0;

	/** フェーズ */
	private		int				phase			= 0;

	/** 時間 */
	private		int				time			= 0;

	/** 回転速度 */
	private		double			s				= 4.0;

	/** 速度Z */
	private		double			mz				= -2.0;

	/** スプライト一覧 */
	private		Group[]			sprites			= null;

	/** 待ち時間一覧 */
	private		int[]			timeWaits		= null;

	/**
	 * 初期化
	 */
	public static void init() {

		try{
			ObjReader.init(TITLE_P , "3d/TitleP.obj");
			ObjReader.init(TITLE_R1, "3d/TitleR.obj");
			ObjReader.init(TITLE_U , "3d/TitleU.obj");
			ObjReader.init(TITLE_S , "3d/TitleS.obj");
			ObjReader.init(TITLE_H , "3d/TitleH.obj");
			ObjReader.init(TITLE_E , "3d/TitleE.obj");
			ObjReader.init(TITLE_R2, "3d/TitleR.obj");
		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end init

	/**
	 * コンストラクタ
	 * @param application PRApplication
	 */
	public PRSpriteTitlePrusher(PRApplication application){

		// スーパークラスを呼び出す
		super(application);

		try{

			// 回転軸
			Point3D p = new Point3D(1.0, 0.0, 0.0);

			// スプライト一覧を生成
			sprites = new Group[COUNT_SPRITES];

			// 待ち時間一覧を生成
			timeWaits = new int[COUNT_SPRITES];

			// マテリアルを生成
			PhongMaterial material = new PhongMaterial();
			material.setDiffuseColor(Color.web("808080"));

			// スプライト一覧の件数分、ループ
			for(int i = 0; i < COUNT_SPRITES; i++){

				// スプライトを生成
				Group sprite = new Group();

				// スプライト一覧に設定
				sprites[i] = sprite;

				// オブジェクト
				HashMap<String, Group> map = ObjReader.create(sprite, TitleS[i]);

				// 回転させる
				sprite.getTransforms().add(INIT_ROTATE);
				sprite.getTransforms().add(INIT_TRANSLATE);

				// 回転軸を設定
				sprite.setRotationAxis(p);

				// メッシュビューを取得
				Group group = map.get("Text");
				MeshView meshView = ((MeshView)group.getChildren().get(0));

				// マテリアルを登録
				meshView.setMaterial(material);

				// グループを登録
				this.getChildren().add(sprite);

				// 待ち時間設定
				timeWaits[i] = 30 * i;

			} // end for スプライト一覧の件数分、ループ

		}catch(Exception ex){
			ex.printStackTrace();
		}

	} // end PRSpriteTitlePrusher

	/**
	 * 実行
	 */
	@Override
	public void run(){

		// フェーズ０
		if(phase == 0){

			// 時間初期化
			time = 0;

			// フェーズ１へ
			phase = 1;

		// フェーズ１
		}else if(phase == 1){

			// スプライト一覧の件数分、ループ
			for(int i = 0; i < COUNT_SPRITES; i++){

				// スプライトを取得
				Group sprite = sprites[i];

				// 座標取得
				double z = sprite.getTranslateZ();

				// 座標加算
				z = z + mz;

				// 定位置まで来ている場合、
				if(z <= Z){
					// Z補正
					z = Z;
				}

				// 座標設定
				sprite.setTranslateZ(z);

			} // end for スプライト一覧の件数分、ループ

			// スプライトを取得
			Group sprite = sprites[COUNT_SPRITES - 1];

			// 座標取得
			double z = sprite.getTranslateZ();

			// 定位置まで来ている場合、
			if(z <= Z){

				// 時間初期化
				time = 0;
				// フェーズ２へ
				phase = 2;

			}

		// フェーズ２
		}else if(phase == 2){

			// 時間＋１
			time++;

			// 時間が経ったら、
			if(time > 60){

				// 時間初期化
				time = 0;
				// フェーズ３へ
				phase = 3;

				// フェーズ設定
				super.application.setPhase(PHASE_TITLE_3);

			}

		// フェーズ３
		}else if(phase == 3){

			// スプライト一覧の件数分、ループ
			for(int i = 0; i < COUNT_SPRITES; i++){

				// スプライトを取得
				Group sprite = sprites[i];

				// 時間が来たら
				if(timeWaits[i] <= 0){

					// 角度取得
					int rotate = (int)sprite.getRotate();

					// 360度以上の場合、
					if(rotate >= 360){

						// 待ち時間を設定
						timeWaits[i] = 180;

						// 回転をデフォルト値へ
						sprite.setRotate(0);

					// 360度未満の場合、
					}else{

						// 回転
						sprite.setRotate(sprite.getRotate() + this.s);

					}

				// 時間が来てない場合、
				}else{
					// 待ち時間を減らす
					timeWaits[i]--;
				}

			} // end for スプライト一覧の件数分、ループ

		}

	} // end run

	/**
	 * 初期化
	 */
	public void initPosition(){

		// フェーズ０
		phase = 0;

		// スプライト一覧の件数分、ループ
		for(int i = 0; i < COUNT_SPRITES; i++){

			// スプライトを取得
			Group sprite = sprites[i];

			// 座標設定
			sprite.setTranslateX(0.5 * i - 1.75);
			sprite.setTranslateY(-1.8);
			sprite.setTranslateZ(200.0 + i * 50);

			// 回転
			sprite.setRotate(0);

			// 待ち時間設定
			timeWaits[i] = 30 * i;

		} // end for スプライト一覧の件数分、ループ

	} // end initPosition

	/**
	 * タイトルをスキップ
	 */
	public void skipTitle(){

		// スプライト一覧の件数分、ループ
		for(int i = 0; i < COUNT_SPRITES; i++){
			// Z設定
			sprites[i].setTranslateZ(Z);
		} // end for スプライト一覧の件数分、ループ

		// 時間初期化
		time = 0;

		// フェーズ２へ
		phase = 2;

	} // end skipTitle

}
