/**
 * ストーリークラス
 * <pre>
 * 時間を管理し、タイミングが来たら、スプライトオブジェクトを生成します。
 * </pre>
 */
public class PRStoryPrusher extends PRStory {

	/**
	 * コンストラクタ
	 * @param panel PRApplication
	 */
	public PRStoryPrusher(PRApplication panel) throws Exception{

		// ファイル名を追加し、スーパークラスを呼び出す
		super(panel, "PRStoryPrusher.txt");

	}

	/**
	 * ストーリー呼び出し
	 * @param storyEvent イベント
	 * @param strParam パラメータ
	 * @return スプライトオブジェクト
	 */
	public Object storyEvent(String storyEvent, String strParam){

		Object spriteObject = null;

		// 木
		if("Tree".equals(storyEvent)){
			spriteObject = new PRSpriteGroupTree(super.getPRApplication(), strParam);

		// 柱
		}else if("Pole".equals(storyEvent)){
			spriteObject = new PRSpriteGroupPole(super.getPRApplication(), strParam);

		// ビル
		}else if("Building".equals(storyEvent)){
			spriteObject = new PRSpriteGroupBuilding(super.getPRApplication(), strParam);

		// 石
		}else if("Stone".equals(storyEvent)){
			spriteObject = new PRSpriteGroupStone(super.getPRApplication());

		// 低木
		}else if("Shrub".equals(storyEvent)){
//			spriteObject = new PRSpriteGroupShrub(super.getPRApplication());

		// 岩
		}else if("Rock".equals(storyEvent)){
			spriteObject = new PRSpriteGroupRock(super.getPRApplication(), strParam);

		// アステロイダー
		}else if("Ship".equals(storyEvent)){
			spriteObject = new PRSpriteGroupShip(super.getPRApplication(), strParam);

		// レッド・アイ
		}else if("RedEye".equals(storyEvent)){
			spriteObject = new PRSpriteGroupRedEye(super.getPRApplication(), strParam);

		// 天狗
		}else if("Tengu".equals(storyEvent)){
			spriteObject = new PRSpriteGroupTengu(super.getPRApplication(), strParam);

		// 雲を開始
		}else if("StartCloud".equals(storyEvent)){
			super.getPRApplication().startCloud();

		// 雲を終了
		}else if("FinishCloud".equals(storyEvent)){
			super.getPRApplication().finishCloud();

		// 雲
		}else if("Cloud".equals(storyEvent)){
			spriteObject = new PRSpriteCloud(super.getPRApplication(), strParam);

		// BGM位置
		}else if("BGM".equals(storyEvent)){
			super.getPRApplication().startMP3BGM(Double.valueOf(strParam));

		// BGMフェードアウト
		}else if("Fadeout".equals(storyEvent)){
			if(strParam == null){
				super.getPRApplication().fadeoutMP3BGM();
			}else{
				super.getPRApplication().fadeoutMP3BGM(Integer.valueOf(strParam));
			}

		// ステージクリア
		}else if("StageClear".equals(storyEvent)){
			super.getPRApplication().setPhase(PRConstant.PHASE_STAGE_CLEAR);

		// ゲームクリア
		}else if("GameClear".equals(storyEvent)){
			super.getPRApplication().setPhase(PRConstant.PHASE_GAME_CLEAR);

		} // end if

		// リターン
		return spriteObject;

	} // end storyEvent

}
